"""
The objects defined in this file are meant to run
some kind of processing on a video in a streaming fashion.

They should have a `process` method that can get called on
a frame or a stack of frame. A frame is a 3d-array of size
(height, width, n_colors), where n_colors is the number of color
channels, usually this is 3 (RGB). A stack of frame has one or more
extra dimensions in the front
"""

import abc
import numpy as np
from collections import deque
from threading import Thread
import queue
import time


class ProcessorBase(abc.ABC):
    """
    Base class for online processing objects

    The derived classes should implement the __process__ method.
    """

    def __init__(self, monitor=False, qlen=10):
        self.monitor = monitor

        self.deltas = deque([], qlen)  # list of time delta between two batch of frames
        self.avg_fps = 0.0

        self.monitor_freq = 0.5  # print new message every second

        self.total_frames = 0
        self.start = time.perf_counter()
        self.runtime = 0.0

        self.queue = queue.Queue()

        self.frames_since_last_call = 0
        self.last_print = self.start

        # attributes related to the threaded processing
        self._is_running = True
        self._thread = Thread(target=self._process_loop, args=())
        self._thread.start()

    def stop(self):
        """ Interface to stop the thread """
        # stop the processing
        self._is_running = False

        # wait for processing loop to terminate
        self._thread.join()

        # do the final processing
        self.__finalize__()

    def process(self, frames):
        """ Interface to add new frames for processing """
        if self._is_running:
            self.queue.put(frames)

    def __len__(self):
        return self.queue.qsize()

    @property
    def fps(self):
        """ Number of frames processed per second """
        return self.total_frames / self.runtime if self.runtime > 0 else 0

    def print_perf(self):
        """ Prints the number of frames processed per second to standard output """
        print(
            "{} frames processed in {} seconds (fps: {:7.3f})".format(
                self.total_frames, self.runtime, self.fps
            )
        )

    @abc.abstractmethod
    def __finalize__(self):
        """
        Abstract method called at the end of the processing
        It can be called to save the data to file, etc
        """
        pass

    @abc.abstractmethod
    def __process__(self, frames):
        """
        Abstract method called to process the frames
        """
        pass

    def _process_loop(self):
        """ The processing loop running in a separate thread """

        while self._is_running:

            try:
                frames = self.queue.get(block=True, timeout=0.1)
            except queue.Empty:
                continue

            # do the work
            self.__process__(frames)

            if self.monitor:

                # now measure the frame rate
                if frames.ndim <= 3:
                    n_frames = 1
                elif frames.ndim > 3:
                    n_frames = np.prod(frames.shape[:-3])

                self.frames_since_last_call += n_frames
                self.total_frames += n_frames

                now = time.perf_counter()

                delta = now - self.last_print
                self.runtime = now - self.start

                if delta > self.monitor_freq:

                    self.deltas.appendleft((delta, self.frames_since_last_call))

                    avg_fps = np.mean([n / d for d, n in self.deltas])

                    print("avg fps: {:7.3f}".format(avg_fps), end="\r")

                    self.last_print = now
                    self.frames_since_last_call = 0


class ReadSpeedMonitor(ProcessorBase):
    """ A dummy class to monitor the video frame read speed """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __process__(self, frames):
        pass

    def __finalize__(self):
        pass


class OnlineStats(ProcessorBase):
    """
    Compute statistics on the input data in an online way

    Parameters
    ----------
    shape: tuple of int
        Shape of a data point tensor

    Attributes
    ----------
    mean: array_like (shape)
        Mean of all input samples
    var: array_like (shape)
        Variance of all input samples
    count: int
        Sample size
    """

    def __init__(self, shape, monitor=False, qlen=10):
        """
        Initialize everything to zero
        """
        # call parent method
        ProcessorBase.__init__(self, monitor=monitor, qlen=qlen)

        self.shape = shape
        self.mean = np.zeros(shape, dtype=np.float64)
        self.var = np.zeros(shape, dtype=np.float64)
        self.count = 0

    def __process__(self, data):
        """
        Update statistics with new data

        Parameters
        ----------
        data: array_like
            A collection of new data points of the correct shape in an array
            where the first dimension goes along the data points
        """

        if data.shape[-len(self.shape) :] != self.shape:
            raise ValueError(
                "The data.shape[1:] should match the statistics object shape"
            )

        data.reshape((-1,) + self.shape)

        count = data.shape[0]
        mean = np.mean(data, axis=0)
        var = np.var(data, axis=0)

        m1 = self.var * (self.count - 1)
        m2 = var * (count - 1)
        M2 = (
            m1
            + m2
            + (self.mean - mean) ** 2 * count * self.count / (count + self.count)
        )

        self.mean = (count * mean + self.count * self.mean) / (count + self.count)
        self.count += count
        self.var = M2 / (self.count - 1)


class BoxCatcher(ProcessorBase):
    """
    This is a simple object that collect the values of a few pixels on the
    stream of frames

    Parameters
    ----------
    pixels: list of tuples
        The location of the pixels to collect in the image
    box_size: list  or tuple of two int
        The width and height of the bounding box to use for averaging
    """

    def __init__(self, pixels, box_size, monitor=False):

        # call parent method
        ProcessorBase.__init__(self, monitor=monitor)

        # set the attributes
        self.pixels = pixels
        self.data = []
        self.box_size = box_size

        # precompute the slices for each pixel
        off_w = self.box_size[0] // 2
        off_h = self.box_size[1] // 2
        self.ranges = [
            [
                slice(loc[0] - off_w, loc[0] - off_w + self.box_size[0]),  # columns
                slice(loc[1] - off_h, loc[1] - off_h + self.box_size[1]),  # row
            ]
            for loc in self.pixels
        ]

    def __process__(self, frames):
        """
        Catch the values of the pixels in a stack of frames

        Parameters
        ----------
        frames: array_like (..., width, height, n_colors)
            The stack of frames
        """
        self.data.append([frames[r_h, r_w].copy() for r_w, r_h in self.ranges])

    def __finalize__(self):
        """
        Consolidate collected data in numpy ndarray with shape (n_frames, n_pixels, box_height, box_width, 
        """
        self.data = np.array(self.data)



# Copyright 2020 Robin Scheibler
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This file defines the routine to read and write files
created with BlinkyViewer


The underlying format is [MessagePack](https://github.com/msgpack/msgpack-python)
which is suitable to efficiently save binary data.


The encoder/decoder were modelled after the
[msgpack-numpy package](https://github.com/lebedov/msgpack-numpy).
"""
import msgpack
import numpy as np
from datetime import datetime
from .version import __version__
from .utils import pixel_to_str


def encoder(obj, chain=None):
    """ Custom encoder to store numpy.ndarray in MessagePack format """

    if isinstance(obj, np.ndarray):

        # Make sure this is not a structured array type
        assert obj.dtype.kind != "V", "Unsupported non-numeric type"

        return {
            "__nd__": True,  # indicate this is a numpy ndarray
            "shape": obj.shape,
            "dtype": obj.dtype.str,
            "data": obj.tobytes(),
        }
    else:
        return obj if chain is None else chain(obj)


def decoder(obj, chain=None):
    """ Custom decoder to recover numpy.ndarray saved in MessagePack format """

    try:
        if "__nd__" in obj:
            return np.frombuffer(obj["data"], dtype=np.dtype(obj["dtype"])).reshape(
                obj["shape"]
            )
        else:
            return obj if chain is None else chain(obj)

    except KeyError:
        print("error!")
        # This is just a security in case an unrelated document
        # contains "__nd__" as a key
        return obj if chain is None else chain(obj)


class BlinkyFile(object):
    def __init__(self, locations, data, fps, version=None, creation=None, **metadata):
        self.locations = locations
        self.data = data

        assert len(self.locations) == self.data.shape[1], (
            "Error when creating the Blinky file object. "
            "The number of locations should correspond to the "
            "number of signals recorded."
        )

        self.fps = fps
        self.version = version if version is not None else __version__
        self.creation = (
            creation
            if creation is not None
            else datetime.now().astimezone().isoformat()
        )
        self.metadata = metadata

    def dump(self, filename):
        """ Saves the object as a MessagePack file """
        with open(filename, "wb") as f:
            msgpack.pack(self.__dict__, f, default=encoder, use_bin_type=True)

    @classmethod
    def load(cls, filename):
        """ Load a BlinkyFile object from MessagePack format """
        with open(filename, "rb") as f:
            content = msgpack.unpack(f, object_hook=decoder, raw=False)
        return cls(**content)


def file_preview():
    """
    Preview a Blinky file
    """
    import matplotlib.pyplot as plt
    import argparse

    parser = argparse.ArgumentParser(description="Preview a Blinky file")
    parser.add_argument("filename", type=str, help="File name")
    args = parser.parse_args()

    bfile = BlinkyFile.load(args.filename)
    data = bfile.data.astype(np.float)

    if len(data.shape) == 5:
        ## This is a color file, we will average the colors
        data = np.mean(data, axis=-1)

    # Now, for the purpose of preview, we average the boxes
    data = np.mean(data, axis=(-2, -1))

    # Make the time axis
    time = np.arange(data.shape[0]) / bfile.fps

    # Make the plot
    fig, ax = plt.subplots(1, 1)
    ax.plot(time, data)
    ax.set_xlabel("Time [s]")
    ax.set_ylabel("Pixel value")
    ax.set_title(f"File creation {bfile.creation}")
    ax.legend([pixel_to_str(p) for p in bfile.locations])

    plt.show()


if __name__ == "__main__":
    file_preview()



# Copyright 2020 Robin Scheibler
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import tkinter
from tkinter import *
from tkinter.scrolledtext import ScrolledText
from tkinter.filedialog import asksaveasfilename

import PIL
import PIL.ImageTk
import numpy as np
import cv2

try:
    import icube_sdk
    icube_sdk_available = True
except ImportError:
    icube_sdk_available = False

from .gui_utils import *
from .video import ThreadedVideoStream
from .processors import ReadSpeedMonitor, BoxCatcher
from .utils import pixel_to_str
from .io import BlinkyFile


def toggle(x: bool):
    return x ^ True


# Now come the GUI part
class BlinkyViewer(object):
    def __init__(self, window, video_source=0, industrial=False):
        self.window = window
        self.window.protocol("WM_DELETE_WINDOW", self.on_closing_callback)
        # overload "Exit" function for Menu file -> quit
        self.window.createcommand("exit", self.on_closing_callback)

        # window update delay
        self.delay = 33  # milliseconds

        # control
        self.checksat = False
        self.convert_bw = False
        self.convert_log = False

        # zoom size
        self.zoom_patch_size = 31

        # open video feed
        self.video_source = video_source
        self.processor = ReadSpeedMonitor(monitor=True)

        # Choose the video source
        self.industrial = industrial

        self.vid = None
        self.start_video()

        # Keep track of the current frame
        self.current_frame = None

        # GEOMETRY #
        ############

        # Here we want to compute the size of the different boxes
        screen_w = self.window.winfo_screenwidth()
        screen_h = self.window.winfo_screenheight()
        vid_h, vid_w = self.vid.shape

        # set the size of the video canvas
        self.vid_can_w = int(screen_w * 0.75)
        self.vid_can_h = int(self.vid_can_w * (vid_h / vid_w))

        if self.vid_can_h > 0.9 * screen_h:
            self.vid_can_h = int(screen_h * 0.75)
            self.vid_can_w = int(self.vid_can_h * (vid_w / vid_h))

        self.window_w = int(1.33 * self.vid_can_w)

        self.zoom_wh = int(0.33 * self.vid_can_w)
        self.fig_w = self.zoom_wh
        self.fig_h = self.vid_can_h - self.zoom_wh
        self.plist_w = self.zoom_wh
        self.plist_h = int(0.9 * screen_h) - self.vid_can_h

        self.buttons_w = int(self.vid_can_w / 3)
        self.buttons_h = int(0.9 * screen_h) - self.vid_can_h

        self.console_w = int(self.vid_can_w / 3)
        self.console_h = int(0.9 * screen_h) - self.vid_can_h

        self.info_w = self.vid_can_w - self.buttons_w - self.console_w
        self.info_h = int(0.9 * screen_h) - self.vid_can_h

        self.window_h = self.vid_can_h + self.plist_h

        self.window_w = int(1.03 * self.window_w)
        self.window_h = int(1.03 * self.window_h)

        self.window.geometry(f"{self.window_w}x{self.window_h}+0+0")

        # THE LEFT PANEL #
        ##################

        # create canvas of the right size
        self.canvas = Canvas(
            self.window, width=self.vid_can_w, height=self.vid_can_h
        )
        self.canvas.grid(row=0, column=0, rowspan=2, columnspan=4)
        self.canvas.bind("<Button-1>", self.mouse_callback)

        # The control frame with all buttons
        self.frame_control = Frame(self.window)
        self.frame_control.grid(row=2, column=0, sticky=N)

        # The Buttons
        self.frame_control_buttons = Frame(self.frame_control)
        self.frame_control_buttons.grid(row=0, column=0)

        self.btn_checksat = Button(
            self.frame_control_buttons, text="Sat", width=10, command=self.toggle_checksat
        )
        self.btn_checksat.grid(row=0, column=0)

        self.btn_convert_bw = Button(
            self.frame_control_buttons, text="BW", width=10, command=self.toggle_convert_bw
        )
        self.btn_convert_bw.grid(row=1, column=0)

        self.btn_convert_log = Button(
            self.frame_control_buttons, text="Log", width=10, command=self.toggle_convert_log
        )
        self.btn_convert_log.grid(row=2, column=0)

        # The processing part
        self.frame_control_proc = Frame(self.window)
        self.frame_control_proc.grid(row=2, column=1, sticky="n")

        self.label_frame = Label(
            self.frame_control_proc, text=f"Record:"
        )
        self.label_frame.grid(row=0, column=0, sticky="nw")

        self.output_filename = DEFAULT_FILENAME
        self.label_file = Label(
            self.frame_control_proc, text=f"Output file:"
        )
        self.label_file.grid(row=1, column=0, sticky="nw")

        self.entry_filename = Entry(self.frame_control_proc, width=20)
        self.entry_filename.insert(0, self.output_filename)
        self.entry_filename.grid(row=1, column=1, sticky="nw")

        self.label_boxsize = Label(self.frame_control_proc, text="Box size:")
        self.label_boxsize.grid(row=2, column=0, sticky="nw")

        self.entry_boxsize = Entry(self.frame_control_proc, width=3)
        self.entry_boxsize.insert(0, "1")
        self.entry_boxsize.grid(row=2, column=1, sticky="nw")

        self.btn_process = Button(
            self.frame_control_proc, text=PROCESS_LABEL, width=10, command=self.process_callback
        )
        self.btn_process.grid(row=3, column=0)

        self.canvas_info = InfoBox({ "fps_video": 0, "fps_proc": 0 }, self.window, width=self.info_w // 2, height=self.info_h)
        self.canvas_info.grid(row=2, column=2, sticky="n")

        # The Console
        self.canvas_console = Canvas(self.window, width=self.console_w, height=self.console_h)
        self.canvas_console.grid(row=2, column=3, sticky="ne")
        self.console = ScrolledText(self.canvas_console, state=DISABLED, borderwidth=2, relief="solid", width=60, height=12)
        self.console.grid(row=0, column=0, sticky="ne")

        # THE RIGHT PANEL #
        ###################

        # The pixel tracking figure
        self.pixel_tracker = PixelTracker(
            row=self.vid.height // 2,
            col=self.vid.width // 2,
            width=self.fig_w,
            height=self.fig_h,
            buffer_size=100,
            master=self.window,
        )
        self.pixel_tracker.grid(row=0, column=4)

        # The zoom
        self.canvas_zoom = ZoomCanvas(
            self.zoom_patch_size,
            self.zoom_wh,
            self.vid.shape,
            self.window,
        )
        self.canvas_zoom.grid(row=1, column=4)
        self.canvas_zoom.bind("<Button-1>", self.zoom_callback)

        # Notify pixel tracker of initial zoom selection
        self.pixel_tracker.add(self.canvas_zoom.selected, label=PREVIEW_LABEL)

        # Buttons to add and remove pixels
        self.frame_pix_btn = Frame(self.window)
        self.frame_pix_btn.grid(row=2, column=4)

        self.btn_add_pixel = Button(
            self.frame_pix_btn,
            text="Add Pixel",
            width=10,
            command=self.add_pixel_callback,
        )
        self.btn_add_pixel.grid(row=0, column=0, sticky="nw")

        self.btn_drop_pixel = Button(
            self.frame_pix_btn,
            text="Drop Pixel",
            width=10,
            command=self.drop_pixel_callback,
        )
        self.btn_drop_pixel.grid(row=0, column=1, sticky="nw")

        # The selected pixels list
        self.pixel_list = PixelList(self.frame_pix_btn, width=self.plist_w // 2, height=self.plist_h)  #, list_width=40, list_height=10)
        self.pixel_list.grid(row=0, column=2, rowspan=2, sticky="nw")

        self.log("Welcome to BlinkyViewer")

        self.update()

        # Populate the window
        window.title("BlinkyViewer")

        self.window.mainloop()

    def start_video(self):
        # first stop the video if already running
        if self.vid is not None:
            self.vid.stop()

        if not self.industrial:
            self.vid = ThreadedVideoStream(self.video_source)
        else:
            if not icube_sdk_available:
                raise ValueError("The Driver for the DN3V camera is not available")

            self.vid = icube_sdk.ICubeCamera(self.video_source, 200)
            self.vid.start()

        print("Video FPS:", self.vid.fps)

    def log(self, message):
        self.console.configure(state="normal")
        self.console.insert("end", "\n" + message)
        self.console.configure(state="disabled")
        self.console.see("end")

    def coord_canvas_to_video(self, col, row):
        """ Transforms from canvas to video coordinate """
        c = int((col / self.vid_can_w) * self.vid.width)
        r = int((row / self.vid_can_h) * self.vid.height)
        return c, r

    def coord_video_to_canvas(self, col, row):
        """ Transforms from video to canvas coordinate """
        c = int((col / self.vid.width) * self.vid_can_w)
        r = int((row / self.vid.height) * self.vid_can_h)
        return c, r

    def update(self):

        fps_proc = self.processor.fps if self.processor is not None else 0.
        self.canvas_info.update(fps_video=f"{self.vid.fps:6.2f}", fps_proc=f"{fps_proc:6.2f}")

        new_frame = None

        # Get a frame from the video source
        while self.vid.available:
            new_frame = np.array(self.vid.read(block=False), copy=False)

            if new_frame is None:
                break

            if self.processor is not None:
                self.processor.process(new_frame)

            self.pixel_tracker.push(new_frame)

        # If the video stream stopped, restart it
        if not self.vid.is_streaming:
            # if recording, stop
            if self.process_is_recording():
                self.process_callback()

            # restart video
            self.start_video()

            self.log("Looping the video")

        if new_frame is not None:
            self.current_frame = new_frame

        if self.current_frame is not None:

            # shorten
            frame = self.current_frame

            self.pixel_tracker.update()

            if frame.ndim == 2:
                frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2RGB)

            elif self.convert_bw:
                # convert to black and white
                frame = cv2.cvtColor(frame, cv2.COLOR_RGB2GRAY)
                # but convert back to color so that we
                # an add visual help in color
                frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2RGB)

            if self.checksat:
                sat_pix = np.where(frame == 255)
                sat_col = (0, 255, 0)
                for col in range(3):
                    sat_pix[2][:] = col
                    frame[sat_pix] = sat_col[col]

            if self.convert_log:
                tmp = frame.astype(np.float32)
                tmp = np.log2(tmp + 1)
                frame = (tmp * 255 / np.max(tmp)).astype(np.uint8)

            # create zoomed-in vignette
            self.canvas_zoom.update(frame)

            # Display the video frame
            frame = cv2.resize(
                frame, (self.vid_can_w, self.vid_can_h), interpolation=cv2.INTER_NEAREST
            )

            # indicate the location of selected pixels
            frame = self.pixel_tracker.mark(frame, conv_func=self.coord_video_to_canvas)
            frame = self.canvas_zoom.mark(frame, conv_func=self.coord_video_to_canvas)

            self.photo = PIL.ImageTk.PhotoImage(image=PIL.Image.fromarray(frame))
            self.canvas.create_image(0, 0, image=self.photo, anchor=NW)

        self.window.after(self.delay, self.update)

    def toggle_checksat(self):
        self.checksat = toggle(self.checksat)

    def toggle_convert_bw(self):
        self.convert_bw = toggle(self.convert_bw)

    def toggle_convert_log(self):
        self.convert_log = toggle(self.convert_log)

    def add_pixel_callback(self):
        self.pixel_list.add(self.canvas_zoom.selected)
        self.pixel_tracker.add(self.canvas_zoom.selected)
        self.log(f"Selected pixel at {pixel_to_str(self.canvas_zoom.selected)}")

    def drop_pixel_callback(self):
        for plbl in self.pixel_list.curselection:
            self.pixel_tracker.remove(plbl)
        self.pixel_list.curdelete()
        self.log(f"Dropped pixel at {pixel_to_str(self.canvas_zoom.selected)}")

    def choose_file_callback(self):
        """ Called when pressing on the choose file button """
        tmp_filename = asksaveasfilename(defaultextension=".blinky", initialdir=".")
        if tmp_filename != "":
            self.output_filename = tmp_filename
            self.label_file.config(text=f"Output file: {self.output_filename}")

    def process_is_recording(self):
        return self.btn_process.cget("text") == STOP_LABEL

    def process_is_not_recording(self):
        return self.btn_process.cget("text") == PROCESS_LABEL

    def process_callback(self):
        """ Called when pressing the "Process" button """

        if self.process_is_not_recording():

            bbox = self.entry_boxsize.get()
            pixel_list = self.pixel_list.get()
            self.output_filename = self.entry_filename.get()

            if self.output_filename == "":
                self.log("Please choose a filename")

            elif not bbox.isdigit():
                self.log("The box size should be a number")

            elif len(pixel_list) == 0:
                self.log("No pixel has been selected for recording.")

            else:

                bbox = int(bbox)

                self.log("Start recording")

                # If the video is from a file, we restart
                if not isinstance(self.video_source, int):
                    self.vid.stop()

                if self.processor is not None:
                    self.processor.stop()

                self.processor = BoxCatcher(
                    pixel_list, [bbox, bbox], monitor=True
                )

                # If the video is from a file, we restart
                if not isinstance(self.video_source, int):
                    self.vid = ThreadedVideoStream(self.video_source)

                self.btn_process.config(text=STOP_LABEL)

        elif self.process_is_recording():

            self.log("Stop recording")

            self.btn_process.config(text=PROCESS_LABEL)

            if isinstance(self.processor, BoxCatcher):
                # stop the processor
                self.processor.stop()

                # save the result in blinky file
                new_file = BlinkyFile(
                    self.processor.pixels, self.processor.data, self.vid.fps
                )
                new_file.dump(self.output_filename)

                # Replace by the simple speed meter
                self.processor = ReadSpeedMonitor(monitor=True)

        else:
            raise ValueError("Invalid button state")

    def mouse_callback(self, event):
        """ Called when clicking on the main camera display """
        col, row = self.coord_canvas_to_video(event.x, event.y)
        frame_shape = self.vid.shape
        if col >= frame_shape[1]:
            col = frame_shape[1] - 1
        if row >= frame_shape[0]:
            row = frame_shape[0] - 1

        self.canvas_zoom.set_origin((col, row))
        self.pixel_tracker.add(self.canvas_zoom.selected, label=PREVIEW_LABEL)

    def zoom_callback(self, event):
        """ Called when clicking on the zoomed out vignette """
        self.canvas_zoom.on_click(event)
        self.pixel_tracker.add(
            self.canvas_zoom.selected, label=PREVIEW_LABEL,
        )

    def on_closing_callback(self):
        self.log("Goodbye!")
        self.vid.stop()
        if self.processor is not None:
            self.processor.stop()

        self.window.destroy()


def start_viewer(video_source, industrial=False):
    BlinkyViewer(Tk(), video_source, industrial=industrial)

def viewer_main():
    import argparse

    parser = argparse.ArgumentParser(
        description="A video viewer to help setup and debug Blinkies"
    )
    parser.add_argument(
        "-v",
        "--video_source",
        type=str,
        default=0,
        help="The video feed number (default 0)",
    )
    parser.add_argument(
        "-i",
        "--industrial",
        action="store_true"
    )
    args = parser.parse_args()

    try:
        video_source = int(args.video_source)
    except ValueError:
        video_source = args.video_source

    start_viewer(video_source, industrial=args.industrial)


if __name__ == "__main__":
    viewer_main()



"""
This file contains routines to help dealing with processing
streaming videos and reading frames and pixel locations

Code written by Daiki Horiike and Robin Scheibler, 2018
"""
import cv2
from threading import Thread
import queue


class FrameGrabber(object):
    def __init__(self):
        self.the_frame = None

    def process(self, frame):
        self.the_frame = frame

    def extract(self):
        return self.the_frame


class ThreadedVideoStream(object):
    """
    This class capture a video (either from live stream or file) using
    the opencv api in a separate thread.
    Parameters
    ----------
    video: int or str
        If an int, this is the index of the video stream. If a str, this is a filename.
    """

    def __init__(self, video, start=0, end=None, qmax_len=200):

        # The maximum number of frames to buffer
        self.qmax_len = qmax_len

        # we'll store frames there
        self.queue = queue.Queue()
        self.video_source = video
        self._start = start

        if end == -1:
            self._end = None
        else:
            self._end = end

        # Try to access the device
        self.capture = cv2.VideoCapture(self.video_source)
        if not self.capture:
            raise ValueError("Couldn" "t open the device.")

        if self._start != 0:
            # So CAP_PROP_POS_FRAMES seems to be broken,
            # per https://github.com/opencv/opencv/issues/9053
            self.pos_frames = self._start // 2  # set first frame to read
            print(
                "Warning: setting start frame seems buggy "
                "(https://github.com/opencv/opencv/issues/9053). "
                "A hack was used. Use at your own risk"
            )

        if self._end is not None:
            self._end = self._end - self._start

        """ Start the stream """
        self._stopped = False  # Use a flag to know the state of the process
        self._count = 0
        self.thread = Thread(target=self._frame_read_loop, args=())
        self.thread.start()

    def __len__(self):
        return self.queue.qsize()

    @property
    def is_streaming(self):
        """ Returns True if capture is running, and False otherwise. """
        return not self._stopped

    @property
    def available(self):
        return self.is_streaming and self.queue.qsize() > 0

    @property
    def width(self):
        return int(self.capture.get(cv2.CAP_PROP_FRAME_WIDTH))

    @property
    def height(self):
        return int(self.capture.get(cv2.CAP_PROP_FRAME_HEIGHT))

    @property
    def brightness(self):
        return self.capture.get(cv2.CAP_PROP_BRIGHTNESS)

    @brightness.setter
    def brightness(self, value):
        self.capture.set(cv2.CAP_PROP_BRIGHTNESS, value)

    @property
    def exposure(self):
        return self.capture.get(cv2.CAP_PROP_EXPOSURE)

    @exposure.setter
    def exposure(self, value):
        self.capture.set(cv2.CAP_PROP_EXPOSURE, value)

    @property
    def auto_exposure(self):
        return self.capture.get(cv2.CAP_PROP_AUTO_EXPOSURE)

    @auto_exposure.setter
    def auto_exposure(self, value):
        self.capture.set(cv2.CAP_PROP_AUTO_EXPOSURE, value)

    @property
    def fps(self):
        return self.capture.get(cv2.CAP_PROP_FPS)

    @property
    def shape(self):
        """ The frame shape in numpy format """
        return (self.height, self.width)

    @property
    def pos_frames(self):
        return self.capture.get(cv2.CAP_PROP_POS_FRAMES)

    @pos_frames.setter
    def pos_frames(self, value):
        self.capture.set(cv2.CAP_PROP_POS_FRAMES, value)

    def __del__(self):
        self.stop()  # stop, just in case

    def stop(self):
        """ Stop the stream """
        self._stopped = True
        self.thread.join()  # wait for frame reading loop to stop
        with self.queue.mutex:
            self.queue.queue.clear()

        # close the video feed
        if self.capture.isOpened():
            self.capture.release()

    def read(self, n=1, block=True, timeout=1):
        """
        Read some frames.
        Parameters
        ----------
        n: int
            Number of frames to retrieve
        block: optional, bool
            Wether to do a blocking call or not
        """

        if self.queue.qsize() == 0 and self._stopped:
            return None

        if n == 1:
            while not self._stopped:
                try:
                    return self.queue.get(block=block, timeout=timeout)
                except queue.Empty:
                    if self._stopped:
                        return None

        elif n > 1:
            ret = []

            while len(ret) < n:
                try:
                    ret.append(self.queue.get(block=block, timeout=timeout))
                except queue.Empty:
                    if self._stopped:
                        return ret

            return ret

        else:
            raise ValueError("n must be strictly positive")

    def _frame_read_loop(self):
        """ This method will fetch the frames in a concurrent thread """

        while not self._stopped:

            if self.queue.qsize() >= self.qmax_len:
                continue

            ret, frame = self.capture.read()

            self._count += 1

            if not ret:
                break
            else:
                # Return RGB frame
                self.queue.put(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
                if self._end is not None and self._count >= self._end:
                    break

        self._stopped = True

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()


def video_stream(video, start=0, end=None, callback=None, show=False):
    """
    Streams a video for display or processing

    Parameters
    ----------
    video: int or str
        If an int, this is the index of the video stream. If a str, this is a filename.
    start: int
        The frame number where to start the streaming
    end: int
        The last frame to stream
    callback: func
        A function to call on each frame
    show: bool
        If True, the video is displayed during streaming
    """

    if show:
        cv2.namedWindow("image", cv2.WINDOW_AUTOSIZE)

    with ThreadedVideoStream(video, start=start, end=end) as cap:

        fps = cap.get_fps()

        while cap.is_streaming:

            frame = cap.read()
            if frame is None:
                break

            if show:
                cv2.imshow("image", frame)
                if cv2.waitKey(1) & 0xFF == ord("q"):
                    break

            if callback is not None:
                callback(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))

    if show:
        cv2.destroyAllWindows()

    return fps


def frame_grabber(video, frame=0, show=False):
    """
    Retrieve a single frame from a video file or stream

    Parameters
    ----------
    video: int or str
        If an int, this is the index of the video stream. If a str, this is a filename.
    frame: int, optional
        The frame index to grab
    show: bool, optional
        Display the frame using matplotlib
    """

    grabber = FrameGrabber()
    video_stream(video, start=frame, end=frame, callback=grabber.process)

    if show:
        import matplotlib.pyplot as plt
        import numpy as np

        plt.imshow(np.array(grabber.extract()))
        plt.show()

    return grabber.extract()


class MouseParam(object):
    """ Helper object to get a mouse click coordinates in an image """

    def __init__(self, input_img_name):
        # parameters of the mouse click
        self.mouse_event = {"x": None, "y": None, "event": None, "flags": None}
        # setting of the mouse callback
        cv2.setMouseCallback(input_img_name, self.__callback, None)

    # callback function
    def __callback(self, event_type, x, y, flags, userdata):

        self.mouse_event["x"] = x
        self.mouse_event["y"] = y
        self.mouse_event["event"] = event_type
        self.mouse_event["flags"] = flags

    def get(self, param_name):
        return self.mouse_event[param_name]

    def get_pos(self):
        return (self.mouse_event["x"], self.mouse_event["y"])


def pixel_from_click(frame):
    """
    Obtain the location of a pixel clicked with the mouse in a frame
    """
    window_name = "input window"
    cv2.imshow(window_name, frame)

    mouse_data = MouseParam(window_name)

    while 1:
        cv2.waitKey(20)
        if mouse_data.get("event") == cv2.EVENT_LBUTTONDOWN:  # left click
            y, x = mouse_data.get_pos()
            break

    cv2.destroyAllWindows()
    return x, y


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(description="Video loader/stream player")
    parser.add_argument("video", default=0, type=str, help="Stream index or file name")
    parser.add_argument(
        "--display",
        "-d",
        action="store_true",
        help="Displays the video as it is streamed",
    )
    args = parser.parse_args()

    try:
        video = int(args.video)
    except:
        video = args.video

    # Start capture
    video_stream(video, show=args.display)



# Copyright 2020 Robin Scheibler
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import numpy as np

blinky_min_db = -80.
blinky_max_db = -10.

blinky_lut = np.array(
    [
        0.0, 0.008773874102714685, 0.017523573381825086, 0.02607612819282401, 0.034620136083068465,
        0.04331733831488338, 0.05230047386528325, 0.061675388827087674, 0.07152582470159385,
        0.08191960702011791, 0.09291522321689072, 0.10456801706542795, 0.11693543765260839,
        0.13008096516534118, 0.1440764950652642, 0.15900309789588218, 0.174950185372309,
        0.19201320590535764, 0.21029006568024444, 0.22987652620980514, 0.2508608672761513,
        0.2733181267338959, 0.29730423713112, 0.32285037488391344, 0.34995782217678406,
        0.37859361622309096, 0.4086872263716097, 0.44012845815276036, 0.4727667360874842,
        0.5064118652974956, 0.5408363170249617, 0.575779026456492, 0.610950634117771,
        0.64604004592588, 0.680722133122291, 0.7146663431266431, 0.747545947214615,
        0.7790476121989043, 0.8088809523456144, 0.8367876959550755, 0.8625500897419622,
        0.8859981647299889, 0.9070155011972668, 0.9255431586351102, 0.9415814810813828,
        0.9551895499249787, 0.9664821367163857, 0.9756241090261105, 0.982822364333689,
        0.9883155116706936, 0.9923616896472224, 0.9952251039284353, 0.9971620885614387,
        0.9984077451488259, 0.9991644930891165, 0.9995941743218074, 0.9998156985914106,
        0.9999105915463267, 0.9999392193803281, 0.999970911572436, 1.0,
    ]
)


def clip(x, lo=0.0, hi=1.0):

    y = x.copy()

    y[y > hi] = hi
    y[y < lo] = lo

    return y


def decibels(x):
    return 10. * np.log10(x)

def decibels_inv(y):
    return 10. ** (y / 10.)


def map_to_unit_interval(x, lo=0., hi=1.):
  """ Linearly map value in [lo_val, hi_val] to [0, 1] """
  return (x - lo) / (hi - lo)

def unmap_from_unit_interval(y, lo=0., hi=1.):
  """ Linearly map value in [0, 1] to [lo_val, hi_val] """
  return y * (hi - lo) + lo


def lut_interp(x, lut):

    x = clip(x, lo=0.0, hi=1.0)
    y = x.copy()

    # bin clipped at 1 should be left alone
    not_1 = y != 1

    # linear interpolation between bins
    p_f = x[not_1] * (len(lut) - 1)
    p = np.floor(p_f).astype(np.int)
    y[not_1] = (p_f - p) * (lut[p + 1] - lut[p]) + lut[p]

    return y


def lut_interp_inv(y, lut):

    y = clip(y, lo=0.0, hi=lut[-1])

    x = np.linspace(0, lut[-1], len(lut))

    y2 = y.flatten()
    output = np.zeros(y2.shape)

    for i in range(y2.shape[0]):
        k = np.where(y2[i] - lut[:-1] >= 0)[0][-1]

        a = (lut[k + 1] - lut[k]) / (x[k + 1] - x[k])
        b = lut[k]

        output[i] = (y2[i] - b) / a + x[k]

    return output.reshape(y.shape)


def blinky_non_linearity(x):
    """
    Non-linear function used in the Blinky

    Parameters
    ----------
    x: numpy.ndarray
        Array of values of signal power
    """
    global blinky_lut

    x = decibels(x)
    x = clip(x, lo=blinky_min_db, hi=blinky_max_db)
    x = map_to_unit_interval(x, lo=blinky_min_db, hi=blinky_max_db)
    x = lut_interp(x, blinky_lut)

    return x


def blinky_non_linearity_inv(y):
    """
    Inverse of the non-linear function used in Blinky

    Parameters
    ----------
    y: numpy.ndarray
        Array of values
    """
    global blinky_lut

    y = lut_interp_inv(y, blinky_lut)
    y = unmap_from_unit_interval(y, lo=blinky_min_db, hi=blinky_max_db)
    y = decibels_inv(y)

    return y



# Copyright 2020 Robin Scheibler
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import numpy as np
from tkinter import *
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure

import PIL
import PIL.ImageTk
import numpy as np
import cv2


PREVIEW_LABEL = "Preview"
PROCESS_LABEL = "Start"
STOP_LABEL = "Stop"
DEFAULT_FILENAME = "myfile.blinky"


def pixel_to_str(pixel):
    return f"({pixel[0]}, {pixel[1]})"


class InfoBox(Frame):
    """
    Parameters
    ----------
    content: dict
        A dictionary of key/val pairs used for labels/content
    *args:
        Positional arguments of TkInter Canvas
    **kwargs:
        Keyword arguments of TkInter Canvas
    """
    def __init__(self, content, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.grid_propagate(False)

        info_label = Label(self, text="Info:")
        info_label.grid(row=0, column=0, columnspan=2, sticky="w")

        self.fields = {}

        for row, (lbl, value) in enumerate(content.items()):
            new = {
                "label": Label(self, text=lbl + ": "),
                "value": Label(self, text=str(value)),
            }
            new["label"].grid(row=row + 1, column=0, sticky="w")
            new["value"].grid(row=row + 1, column=1, sticky="w")
            self.fields[lbl] = new

    def update(self, **content):
        for lbl, value in content.items():
            if lbl in self.fields:
                self.fields[lbl]["value"].config(text=str(value))

class PixelList(Frame):
    def __init__(self, window, *args, **kwargs):

        super().__init__(window, *args, **kwargs)
        self.grid_propagate(False)
        self.pixels = {}

        self.list = Listbox(self, width=25, height=12)  # , width=list_width, height=list_height)
        self.list.grid(row=0, column=0, sticky="nw")

    def add(self, pixel):
        """ Add a pixel to the list """
        label = pixel_to_str(pixel)
        if label not in self.pixels:
            self.pixels[label] = pixel
            self.list.insert(END, label)

    def get(self):
        """ Return the list of pixels """
        return list(self.pixels.values())

    @property
    def curselection(self):
        """ Return label of selected pixel """
        return [self.list.get(p) for p in self.list.curselection()]

    def curdelete(self):
        """ Delete currently selected item """
        cur_selected = self.list.curselection()
        for p in cur_selected:
            p_lbl = self.list.get(p)  # get label
            self.pixels.pop(p_lbl)  # remove from pixel list
            self.list.delete(p)  # delete from listbox


class ZoomCanvas(Canvas):
    def __init__(self, patch_size, win_size, frame_shape, *args, **kwargs):
        self.win_size = win_size
        self.patch_size = patch_size
        self.pixel_size = self.win_size / self.patch_size
        self.frame_shape = frame_shape

        self._n_offset = (self.patch_size - 1) // 2
        self.select_row = self._n_offset + 1
        self.select_col = self._n_offset + 1

        # initialize origin in upper left corner
        self.origin = (self._n_offset, self._n_offset)
        self.selected_local = (0, 0)

        self._vignette = None

        super().__init__(width=self.win_size, height=self.win_size, *args, **kwargs)

        self.image = None

    @property
    def selected(self):
        return (
            self.origin[0] + self.selected_local[0],
            self.origin[1] + self.selected_local[1],
        )

    def on_click(self, event):

        c, r = event.x, event.y

        self.selected_local = (
            round(c / self.pixel_size - 0.5) - self._n_offset,
            round(r / self.pixel_size - 0.5) - self._n_offset,
        )

    def set_origin(self, pixel):

        self.origin = pixel
        self._sanitize_origin()

        # reset local selection upon re-framing
        self.selected_local = (0, 0)

    def _sanitize_origin(self):

        tmp = list(self.origin)

        # Force vignette boundaries to be in the frame
        for i in [0, 1]:
            if tmp[i] - self._n_offset < 0:
                tmp[i] = self._n_offset
            elif tmp[i] - self._n_offset + self.patch_size > self.frame_shape[1 - i]:
                tmp[i] = self.frame_shape[1 - i] - self.patch_size + self._n_offset

        self.origin = tuple(tmp)

    def update(self, frame):

        col = self.origin[0] - self._n_offset
        row = self.origin[1] - self._n_offset

        cut_out = frame[
            row : row + self.patch_size, col : col + self.patch_size,
        ].copy()

        # Resize the patch
        z = cv2.resize(
            cut_out, (self.win_size, self.win_size), interpolation=cv2.INTER_NEAREST
        )

        # Draw a red rectangle around the selected pixel
        p_s = (
            round((self.selected_local[0] + self._n_offset) * self.pixel_size),
            round((self.selected_local[1] + self._n_offset) * self.pixel_size),
        )
        p_e = (p_s[0] + round(self.pixel_size), p_s[1] + round(self.pixel_size))
        z = cv2.rectangle(z, p_s, p_e, (255.0, 0.0, 0.0), 1)

        self._vignette = PIL.ImageTk.PhotoImage(image=PIL.Image.fromarray(z))
        self.create_image(0, 0, image=self._vignette, anchor="nw", tags="zoom_image")

    def mark(self, frame, conv_func=None):
        """ Mark location of zoom in original frame """
        c1 = self.origin[0] - self._n_offset
        r1 = self.origin[1] - self._n_offset
        c2 = c1 + self.patch_size
        r2 = r1 + self.patch_size
        if conv_func is not None:
            c1, r1 = conv_func(c1, r1)
            c2, r2 = conv_func(c2, r2)
        frame = cv2.rectangle(frame, (c1, r1), (c2, r2), (255.0, 0.0, 0.0), 1,)

        return frame


class PixelTracker(object):
    def __init__(
        self,
        row=0,
        col=0,
        buffer_size=100,
        width=500,
        height=400,
        dpi=100,
        master=None,
    ):

        self.width = width
        self.height = height
        self.dpi = dpi

        self.fig = Figure(
            figsize=(self.width / self.dpi, self.height / self.dpi), dpi=self.dpi
        )
        self.fig.set_tight_layout(True)
        self.ax = self.fig.add_subplot(111)
        self.plt_canvas = FigureCanvasTkAgg(self.fig, master=master)
        self.plt_canvas.draw()

        # Attributes for the pixel tracker
        self.buffer_size = buffer_size
        self.reset(row, col)

        self.pixels = {}

    def add(self, pixel, label=None):
        if label is None:
            label = pixel_to_str(pixel)

        # If this is not a preview and the label is already
        # in the list, we abort
        if label != PREVIEW_LABEL and label in self.pixels:
            return

        # Add the pixel
        self.pixels[label] = {
            "loc": pixel,
            "values": [0 for i in range(self.buffer_size)],
        }

    def remove(self, pixel_str):
        self.pixels.pop(pixel_str)

    def grid(self, *args, **kwargs):
        self.plt_canvas.get_tk_widget().grid(*args, **kwargs)

    def reset(self, row, col):
        self.row = row
        self.col = col
        self.values = [0 for i in range(self.buffer_size)]

    def push(self, frame):

        for lbl, pxl in self.pixels.items():

            col, row = pxl["loc"]

            # Make the value grayscale if it isn't already
            if frame.ndim == 3:
                value = cv2.cvtColor(
                    frame[row, col, None, None, :], cv2.COLOR_RGB2GRAY
                )[0, 0]
            elif frame.ndim == 2:
                value = frame[row, col]
            else:
                raise ValueError("Bad input")

            # add to the buffer
            pxl["values"].pop(0)
            pxl["values"].append(value)

    def update(self):

        # update the graph
        self.ax.clear()

        for lbl, pxl in self.pixels.items():
            self.ax.plot(
                np.arange(-len(pxl["values"]) + 1, 1), pxl["values"], label=lbl
            )

        if len(self.pixels) > 0:
            self.ax.legend(loc="lower left")
        self.ax.set_ylim([0, 255])
        self.ax.set_xticks([])
        self.ax.set_yticks([])
        self.fig.tight_layout(pad=0)
        self.plt_canvas.draw()

    def mark(self, frame, conv_func):
        """ Mark the locations of pixels with circles on a frame """
        mark_radius = 4
        mark_color_preview = (0, 0, 255)
        mark_color_selected = (255, 0, 0)
        for lbl, p in self.pixels.items():
            if lbl == PREVIEW_LABEL:
                c = mark_color_preview
            else:
                c = mark_color_selected
            if conv_func is not None:
                loc = conv_func(*p["loc"])
            else:
                loc = p["loc"]
            frame = cv2.circle(frame, loc, mark_radius, c, 2)

        return frame




import setuptools

package_name = "blinkytools"

# import description from readme file
with open("README.md", "r") as fh:
    long_description = fh.read()

# import version from package internal variable
with open(package_name + "/version.py") as f:
    exec(f.read())

setuptools.setup(
    name=package_name,
    version=__version__,
    author="Robin Scheibler",
    author_email="fakufaku@gmail.com",
    description="Software tools to work with Blinky sound-to-light conversion sensors.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/onolab-tmu/blinky",
    install_requires=["numpy", "matplotlib", "msgpack", "opencv-python", "pillow"],
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
